﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using CustomersManager.SL.RemoteCustomersService;
using CustomersManager.SL.ViewModel;
using GalaSoft.MvvmLight.Threading;

namespace CustomersManager.SL.Model
{
    public class CustomersService : ICustomersService
    {
        private const string SaveUriFormat = "http://localhost:55271/SavePictureService.ashx?account={0}";
        private const string PicturePath = "http://localhost:55271/pics/{0}.png";

        public void GetCustomers(
            Action<IEnumerable<CustomerViewModel>, Exception> callback)
        {
            var client = new CustomerServiceClient("BasicEndPoint");
            client.GetAllCustomersCompleted += (s, e) =>
            {
                var userCallback = e.UserState as Action<IEnumerable<CustomerViewModel>, Exception>;

                if (userCallback == null)
                {
                    return;
                }

                if (e.Error != null)
                {
                    userCallback(null, e.Error);
                    return;
                }

                var customers = new List<CustomerViewModel>();

                foreach (var model in e.Result)
                {
                    // Make sure that the picture URI is unique to force loading
                    // the picture from site (to avoid cache problems)
                    model.PictureUri = new Uri(model.PictureUri.AbsoluteUri + "?" + DateTime.Now.Ticks);
                    var customer = new CustomerViewModel(model);
                    customers.Add(customer);
                }

                userCallback(customers.AsEnumerable(), null);
            };

            client.GetAllCustomersAsync(callback);
        }

        public void SaveCustomer(
            CustomerViewModel customer,
            Action<bool> callback)
        {
            var client = new CustomerServiceClient("BasicEndPoint");
            client.SaveCustomerCompleted += (s, e) =>
            {
                var userCallback = e.UserState as Action<bool>;

                if (userCallback == null)
                {
                    return;
                }

                if (e.Error == null)
                {
                    userCallback(e.Result);
                    return;
                }

                userCallback(false);
            };

            client.SaveCustomerAsync(customer.Model, callback);
        }

        public void SavePicture(Stream pictureStream, int accountNumber, Action<bool, string> callback)
        {
            var uriString = string.Format(
                SaveUriFormat,
                accountNumber);
            var uri = new Uri(uriString);

            var request = WebRequest.Create(uri) as HttpWebRequest;
            request.Method = "POST";

            var memoryStream = new MemoryStream();
            CopyToStream(pictureStream, memoryStream);

            var stateElements = new RequestElements
            {
                Callback = callback,
                Request = request,
                ImageStream = memoryStream
            };

            request.BeginGetRequestStream(ReadCompleted, stateElements);
        }

        public CustomerViewModel CreateCustomer(int accountNumber)
        {
            var model = new Customer
            {
                AccountNumber = accountNumber,
                PictureUri = new Uri(string.Format(PicturePath, "unknown"))
            };

            var newCustomer = new CustomerViewModel(model)
            {
                IsDirty = true
            };

            return newCustomer;
        }

        public Uri GetKnownPictureUri(int accountNumber)
        {
            return new Uri(string.Format(PicturePath, accountNumber));
        }

        private static void CopyToStream(Stream input, Stream output)
        {
            int read;
            var buffer = new byte[32768];
            while ((read = input.Read(buffer, 0, buffer.Length)) > 0)
            {
                output.Write(buffer, 0, read);
            }
        }

        private static void ReadCompleted(IAsyncResult result)
        {
            var elements = (RequestElements)result.AsyncState;

            using (var outputStream = elements.Request.EndGetRequestStream(result))
            {
                elements.ImageStream.Seek(0, SeekOrigin.Begin);
                CopyToStream(elements.ImageStream, outputStream);
                elements.ImageStream.Close();
            }

            elements.Request.BeginGetResponse(ResponseCompleted, elements);
        }

        private static void ResponseCompleted(IAsyncResult result)
        {
            var elements = (RequestElements)result.AsyncState;

            var response = elements.Request.EndGetResponse(result) as HttpWebResponse;

            string returnedResult;
            using (var stream = response.GetResponseStream())
            {
                using (var reader = new StreamReader(stream))
                {
                    returnedResult = reader.ReadToEnd();
                }
            }

            if (elements.Callback != null)
            {
                // The response comes on the worker thread ==> Must dispatch to UI thread
                if (returnedResult == "Success")
                {
                    DispatcherHelper.CheckBeginInvokeOnUI(() => elements.Callback(true, string.Empty));
                }
                else
                {
                    DispatcherHelper.CheckBeginInvokeOnUI(() => elements.Callback(false, returnedResult));
                }
            }
        }

        public void RegisterUriForNotifications(Uri notificationUri)
        {
            var client = new CustomerServiceClient("BasicEndPoint");
        }

        public struct RequestElements
        {
            public Action<bool, string> Callback;
            public HttpWebRequest Request;
            public Stream ImageStream;
        }
    }
}
