﻿using System;
using System.Collections.ObjectModel;
using System.Linq;
using CustomersManager.SL.Model;
using CustomersManager.SL.RemoteCustomersService;
using GalaSoft.MvvmLight;
using GalaSoft.MvvmLight.Command;

namespace CustomersManager.SL.ViewModel
{
    public partial class MainViewModel : ViewModelBase
    {
        public const string CustomersPropertyName = "Customers";
        private ObservableCollection<CustomerViewModel> _customers;
        public ObservableCollection<CustomerViewModel> Customers
        {
            get
            {
                return _customers;
            }

            set
            {
                if (_customers == value)
                {
                    return;
                }

                _customers = value;
                RaisePropertyChanged(CustomersPropertyName);
            }
        }

        public const string IsBusyPropertyName = "IsBusy";
        private bool _isBusy;
        public bool IsBusy
        {
            get
            {
                return _isBusy;
            }

            set
            {
                if (_isBusy == value)
                {
                    return;
                }

                _isBusy = value;
                RaisePropertyChanged(IsBusyPropertyName);
            }
        }

        public const string SelectedCustomerPropertyName = "SelectedCustomer";
        private CustomerViewModel _selectedCustomer;
        public CustomerViewModel SelectedCustomer
        {
            get
            {
                return _selectedCustomer;
            }

            set
            {
                if (_selectedCustomer == value)
                {
                    return;
                }

                _selectedCustomer = value;
                RaisePropertyChanged(SelectedCustomerPropertyName);
                UploadPictureCommand.RaiseCanExecuteChanged();
                SaveSelectedCustomerCommand.RaiseCanExecuteChanged();
            }
        }

        public CustomerGender[] Genders
        {
            get;
            private set;
        }

        public IDialogService DialogService
        {
            get;
            internal set;
        }

        public RelayCommand SaveSelectedCustomerCommand
        {
            get;
            private set;
        }

        public RelayCommand UploadPictureCommand
        {
            get;
            private set;
        }

        public RelayCommand RefreshCommand
        {
            get;
            private set;
        }

        public RelayCommand CreateCustomerCommand
        {
            get;
            private set;
        }

        private readonly ICustomersService _customerService;

        public MainViewModel(ICustomersService customerService)
        {
            _customerService = customerService;

            Genders = new[]
            {
                CustomerGender.Male,
                CustomerGender.Female,
                CustomerGender.Boy,
                CustomerGender.Girl
            };

            Refresh();

            SaveSelectedCustomerCommand = new RelayCommand(
                SaveSelectedCustomer,
                () => SelectedCustomer != null);

            UploadPictureCommand = new RelayCommand(
                UploadPicture,
                () => SelectedCustomer != null && !SelectedCustomer.IsNew);

            RefreshCommand = new RelayCommand(Refresh);
            CreateCustomerCommand = new RelayCommand(CreateCustomer);
        }

        private void SaveSelectedCustomer()
        {
            if (!IsBusy
                && SelectedCustomer.IsDirty)
            {
                IsBusy = true;

                _customerService.SaveCustomer(
                    SelectedCustomer, 
                    result =>
                    {
                        if (result)
                        {
                            SelectedCustomer.IsDirty = false;
                            SelectedCustomer.IsNew = false;
                        }
                        else
                        {
                            if (DialogService != null)
                            {
                                DialogService.ShowMessage("Error when saving", "Error");
                            }
                        }

                        IsBusy = false;
                        UploadPictureCommand.RaiseCanExecuteChanged();
                    });
            }
        }
        
        private void UploadPicture()
        {
            if (DialogService == null)
            {
                return;
            }

            var file = DialogService.OpenFile("PNG files (*.png) | *.png");

            if (file != null)
            {
                using (var imageStream = file.OpenRead())
                {
                    _customerService.SavePicture(imageStream, SelectedCustomer.Model.AccountNumber, UploadPictureCompleted);
                }
            }
        }

        private void UploadPictureCompleted(bool success, string errorMessage)
        {
            if (success)
            {
                // Force refresh picture
                var isDirty = SelectedCustomer.IsDirty;
                var uri = _customerService.GetKnownPictureUri(SelectedCustomer.Model.AccountNumber);
                SelectedCustomer.Model.PictureUri = new Uri(uri.AbsoluteUri + "?" + DateTime.Now.Ticks);
                SelectedCustomer.IsDirty = isDirty;
            }
            else
            {
                if (DialogService != null)
                {
                    DialogService.ShowMessage(errorMessage, "Error");
                }
                return;
            }
        }

        private void CreateCustomer()
        {
            var random = new Random();
            var createOtherNumber = false;
            int number;

            do
            {
                number = random.Next(1000, 9999);
                foreach (var customer in Customers)
                {
                    createOtherNumber = (customer.Model.AccountNumber == number);
                    if (createOtherNumber)
                    {
                        break;
                    }
                }
            }
            while (createOtherNumber);

            var newCustomer = _customerService.CreateCustomer(number);
            Customers.Add(newCustomer);
            SelectedCustomer = newCustomer;
        }

        private void Refresh()
        {
            if (IsBusy)
            {
                return;
            }

            IsBusy = true;

            _customerService.GetCustomers((customers, error) =>
            {
                if (error != null)
                {
                    if (DialogService != null)
                    {
                        DialogService.ShowMessage(error.Message, "Error");
                    }
                    return;
                }

#if WINDOWS_PHONE
                Customers = new ObservableCollection<CustomerViewModel>();
                foreach (var customer in customers)
                {
                    Customers.Add(customer);
                }
#else
                Customers = new ObservableCollection<CustomerViewModel>(customers);
#endif

                IsBusy = false;
            });
        }

        ////public override void Cleanup()
        ////{
        ////    // Clean up if needed

        ////    base.Cleanup();
        ////}
    }
}