﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.Serialization;
using System.ServiceModel;
using System.Text;

namespace DuplexPollingSample.Web
{
    [ServiceBehavior(
        InstanceContextMode = InstanceContextMode.PerCall)]
    public class FilesChangedService : IFilesChangedService
    {
        private static FileSystemWatcher _watcher;
        private static List<IFilesChangedClient> _clients;

        private const string WatchedDirectory 
            = "c:\\temp\\DuplexPollingSampleFiles";

        public void StartObservingFiles()
        {
            if (_clients == null)
            {
                _clients = new List<IFilesChangedClient>();
            }

            var client = OperationContext.Current
                .GetCallbackChannel<IFilesChangedClient>();

            // Initial call
            var dir = new DirectoryInfo(WatchedDirectory);
            var files = dir.GetFiles();

            foreach (var file in files)
            {
                client.FileAdded(file.Name.ToLower());
            }

            if (!_clients.Contains(client))
            {
                _clients.Add(client);
            }

            if (_watcher == null)
            {
                _watcher = new FileSystemWatcher(WatchedDirectory);
                _watcher.Created += OnWatcherCreated;
                _watcher.Deleted += OnWatcherDeleted;
                _watcher.Renamed += OnWatcherRenamed;
                _watcher.EnableRaisingEvents = true;
            }
        }

        private void OnWatcherCreated(object sender, FileSystemEventArgs e)
        {
            var toRemove = new List<IFilesChangedClient>();

            foreach (var client in _clients)
            {
                try
                {
                    client.FileAdded(e.Name.ToLower());
                }
                catch (TimeoutException)
                {
                    toRemove.Add(client);
                }
            }

            foreach (var client in toRemove)
            {
                _clients.Remove(client);
            }
        }

        private void OnWatcherRenamed(object sender, RenamedEventArgs e)
        {
            var toRemove = new List<IFilesChangedClient>();

            foreach (var client in _clients)
            {
                try
                {
                    client.FileRenamed(e.OldName.ToLower(), e.Name.ToLower());
                }
                catch (TimeoutException)
                {
                    toRemove.Add(client);
                }
            }

            foreach (var client in toRemove)
            {
                _clients.Remove(client);
            }
        }

        private void OnWatcherDeleted(object sender, FileSystemEventArgs e)
        {
            var toRemove = new List<IFilesChangedClient>();

            foreach (var client in _clients)
            {
                try
                {
                    client.FileDeleted(e.Name.ToLower());
                }
                catch (TimeoutException)
                {
                    toRemove.Add(client);
                }
            }

            foreach (var client in toRemove)
            {
                _clients.Remove(client);
            }
        }

        public void StopObservingFiles()
        {
            var client = OperationContext.Current
                .GetCallbackChannel<IFilesChangedClient>();

            if (_clients != null
                && _clients.Contains(client))
            {
                _clients.Remove(client);
            }
        }
    }
}
