﻿using System;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.IO;
using System.Windows;
using System.Windows.Media;
using GalaSoft.MvvmLight.Command;
using GreetingCardMaker.Helpers;
using System.Windows.Media.Imaging;

namespace GreetingCardMaker.ViewModel
{
    public class MainViewModel : INotifyPropertyChanged
    {
        public event PropertyChangedEventHandler PropertyChanged;

        protected void RaisePropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
            {
                PropertyChanged(this, 
                    new PropertyChangedEventArgs(propertyName));
            }
        }

        internal IDialogService DialogService
        {
            get;
            set;
        }

        public CaptureSource VideoAndAudioSource
        {
            get;
            private set;
        }

        public const string SelectedWebCamPropertyName = "SelectedWebCam";
        private VideoCaptureDevice _webcam;
        public VideoCaptureDevice SelectedWebCam
        {
            get
            {
                return _webcam;
            }

            set
            {
                if (_webcam == value)
                {
                    return;
                }

                if (VideoAndAudioSource.State == CaptureState.Started)
                {
                    VideoAndAudioSource.Stop();
                }

                _webcam = value;
                RaisePropertyChanged(SelectedWebCamPropertyName);

                if (_webcam != null)
                {
                    VideoAndAudioSource.VideoCaptureDevice = _webcam;

                    if ((Microphones.Count == 0
                        || SelectedMicrophone != null)
                        && (CaptureDeviceConfiguration.AllowedDeviceAccess
                        || CaptureDeviceConfiguration.RequestDeviceAccess()))
                    {
                        try
                        {
                            VideoAndAudioSource.Start();
                            StartAudioCommand.RaiseCanExecuteChanged();
                        }
                        catch (InvalidOperationException)
                        {
                            DialogService.ShowMessage("Impossible to start the device");
                        }
                    }
                }
            }
        }

        public const string SelectedMicrophonePropertyName = "SelectedMicrophone";
        private AudioCaptureDevice _microphone;
        public AudioCaptureDevice SelectedMicrophone
        {
            get
            {
                return _microphone;
            }

            set
            {
                if (_microphone == value)
                {
                    return;
                }

                if (VideoAndAudioSource.State == CaptureState.Started)
                {
                    VideoAndAudioSource.Stop();
                }

                _microphone = value;
                RaisePropertyChanged(SelectedMicrophonePropertyName);

                if (_microphone != null)
                {
                    VideoAndAudioSource.AudioCaptureDevice = _microphone;

                    if ((WebCams.Count == 0
                        || SelectedWebCam != null)
                        && (CaptureDeviceConfiguration.AllowedDeviceAccess
                        || CaptureDeviceConfiguration.RequestDeviceAccess()))
                    {
                        try
                        {
                            VideoAndAudioSource.Start();
                            StartAudioCommand.RaiseCanExecuteChanged();
                        }
                        catch (InvalidOperationException)
                        {
                            DialogService.ShowMessage("Impossible to start the device");
                        }
                    }
                }
            }
        }

        public ReadOnlyCollection<VideoCaptureDevice> WebCams
        {
            get;
            private set;
        }

        public ReadOnlyCollection<AudioCaptureDevice> Microphones
        {
            get;
            private set;
        }

        public RelayCommand StartAudioCommand
        {
            get;
            private set;
        }

        public RelayCommand StopAudioCommand
        {
            get;
            private set;
        }

        public RelayCommand<UIElement> CaptureImageCommand
        {
            get;
            private set;
        }

        private WavAudioSink _sink;

        public MainViewModel()
        {
            WebCams = CaptureDeviceConfiguration
                .GetAvailableVideoCaptureDevices();
            Microphones = CaptureDeviceConfiguration
                .GetAvailableAudioCaptureDevices();

            VideoAndAudioSource = new CaptureSource();

            StartAudioCommand = new RelayCommand(
                StartAudio,
                () => SelectedMicrophone != null
                    && _sink == null);

            StopAudioCommand = new RelayCommand(
                StopAudio,
                () => _sink != null);

            CaptureImageCommand = new RelayCommand<UIElement>(CaptureImage);
        }

        private void StartAudio()
        {
            _sink = new WavAudioSink
            {
                CaptureSource = VideoAndAudioSource
            };

            StartAudioCommand.RaiseCanExecuteChanged();
            StopAudioCommand.RaiseCanExecuteChanged();
        }

        private void StopAudio()
        {
            try
            {
                using (var outputStream = DialogService.GetFile(
                    "WAV Files (*.wav) | *.wav", ".wav"))
                {
                    _sink.Save(outputStream);
                }
            }
            catch (IOException ex)
            {
                DialogService.ShowMessage(ex.Message);
            }

            _sink = null;
            StartAudioCommand.RaiseCanExecuteChanged();
            StopAudioCommand.RaiseCanExecuteChanged();
        }

        private void CaptureImage(UIElement element)
        {
            try
            {
                var bitmap = new WriteableBitmap(element, null);

                using (var outputStream = DialogService.GetFile("PNG Files (*.png) | *.png", ".png"))
                {
                    if (outputStream == null)
                    {
                        return;
                    }

                    PngManager.SaveToImage(bitmap, outputStream);
                }
            }
            catch (IOException ex)
            {
                DialogService.ShowMessage(ex.Message);
            }
        }
    }
}
