﻿using System;
using System.Collections.Generic;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.ComponentModel;
using System.Collections.ObjectModel;
using MyNorthwind.Web;
using System.ServiceModel.DomainServices.Client;
using GalaSoft.MvvmLight.Command;
using System.Linq;
using System.Windows.Data;

namespace MyNorthwind.Silverlight.ViewModel
{
    public class MainViewModel : INotifyPropertyChanged
    {
        public const string IsBusyPropertyName = "IsBusy";
        private bool _isBusy = false;

        public bool IsBusy
        {
            get { return _isBusy; }

            set
            {
                if (_isBusy == value)
                {
                    return;
                }

                _isBusy = value;
                RaisePropertyChanged(IsBusyPropertyName);
            }
        }

        private PagedCollectionView _orders = null;
        public PagedCollectionView Orders
        {
            get
            {
                if (_orders == null)
                {
                    _orders = new PagedCollectionView(_context.Orders);
                    _orders.SortDescriptions.Add(
                        new SortDescription(
                            "ShipCity", ListSortDirection.Ascending));
                }
                return _orders;
            }
        }

        private MyNorthwindDomainContext _context;
        private EntityQuery<Order> _currentQuery;
        private int _numberOfOrders = -1;
        private int _pageNumber = 0;
        private const int PageSize = 20;

        public RelayCommand SaveCommand
        {
            get;
            private set;
        }

        public RelayCommand AddCommand
        {
            get;
            private set;
        }

        public RelayCommand<Order> DeleteOrderCommand
        {
            get;
            private set;
        }

        public IDialogService DialogService
        {
            get;
            set;
        }

        public void ShowMessage(string message)
        {
            if (DialogService != null)
            {
                DialogService.ShowMessage(message);
            }
        }

        private bool AskConfirmation(string message)
        {
            if (DialogService != null)
            {
                return DialogService.AskConfirmation(message);
            }
            return true;
        }

        public MainViewModel()
        {
            if (DesignerProperties.IsInDesignTool)
            {
                var list = new List<Order>
                {
                    new Order { OrderID = 1234, ShipCity = "Zurich", Freight = 123 },
                    new Order { OrderID = 3456, ShipCity = "Lausanne", Freight = 42 },
                    new Order { OrderID = 4567, ShipCity = "Basel", Freight = 56 }
                };

                _orders = new PagedCollectionView(list);
            }
            else
            {
                SaveCommand = new RelayCommand(SaveChanges);
                AddCommand = new RelayCommand(AddOrder);
                DeleteOrderCommand = new RelayCommand<Order>(DeleteOrder);
                
                _context = new MyNorthwindDomainContext();
                LoadAllOrders();
            }
        }

        public void LoadAllOrders()
        {
            IsBusy = true;
            _pageNumber = 0;
            _context.Orders.Clear();
            _currentQuery = _context.GetOrdersQuery().OrderBy(o => o.ShipCity);
            _context.Load(_currentQuery, LoadAllOrdersCompleted, null);
        }

        public void LoadAllOrdersByCityAndShip(string cityStart)
        {
            IsBusy = true;
            _pageNumber = 0;
            var query = from o in _context.GetOrdersByCityQuery(cityStart)
                        where o.ShipName.StartsWith("V")
                        orderby o.ShipCity
                        orderby o.OrderID
                        select o;

            _currentQuery = query;
            _context.Load(query, LoadAllOrdersCompleted, null);
            Orders.Filter = o =>
            {
                var order = o as Order;
                return order.ShipCity.StartsWith(cityStart)
                    && order.ShipName != null
                    && order.ShipName.StartsWith("V");
            };
        }

        private void LoadAllOrdersCompleted(LoadOperation<Order> op)
        {
            IsBusy = false;
        }

        public void SaveChanges()
        {
            foreach (var entity in _context.Orders)
            {
                if (entity.HasChanges
                    && entity.HasValidationErrors)
                {
                    ShowMessage("Impossible to save");
                    return;
                }
            }

            IsBusy = true;
            _context.SubmitChanges(SaveChangesCompleted, null);
        }

        private void SaveChangesCompleted(SubmitOperation op)
        {
            IsBusy = false;
        }

        private void AddOrder()
        {
            var lastOrder = _context.Orders.OrderBy(o => o.OrderID).Last();

            var newOrder = new Order
            {
                CustomerID = "ALFKI",
                EmployeeID = 1,
                Freight = 10,
                OrderID = lastOrder.OrderID + 1
            };

            _context.Orders.Add(newOrder);
            _numberOfOrders++;
        }

        private void DeleteOrder(Order order)
        {
            if (!_context.Orders.Contains(order))
            {
                return;
            }

            var confirmation = string.Format(
                "Do you want to delete the order # {0}?",
                order.OrderID);

            if (AskConfirmation(confirmation))
            {
                IsBusy = true;
                _context.Orders.Remove(order);
                _context.SubmitChanges(DeleteOrderCompleted, order);
            }
        }

        private void DeleteOrderCompleted(SubmitOperation op)
        {
            _numberOfOrders--;
            IsBusy = false;
        }

        public event PropertyChangedEventHandler PropertyChanged;

        public void RaisePropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
            {
                PropertyChanged(this,
                    new PropertyChangedEventArgs(propertyName));
            }
        }
    }
}
