﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.ComponentModel;
using System.Collections.ObjectModel;
using MyNorthwind.Web;
using System.ServiceModel.DomainServices.Client;
using GalaSoft.MvvmLight.Command;
using System.Linq;

namespace MyNorthwind.Silverlight.ViewModel
{
    public class MainViewModel : INotifyPropertyChanged
    {
        public const string IsBusyPropertyName = "IsBusy";
        private bool _isBusy = false;

        public bool IsBusy
        {
            get { return _isBusy; }

            set
            {
                if (_isBusy == value)
                {
                    return;
                }

                _isBusy = value;
                RaisePropertyChanged(IsBusyPropertyName);
            }
        }

        public const string OrdersPropertyName = "Orders";
        private ObservableCollection<Order> _orders = null;

        public ObservableCollection<Order> Orders
        {
            get { return _orders; }
            set
            {
                if (_orders == value)
                {
                    return;
                }

                _orders = value;
                RaisePropertyChanged(OrdersPropertyName);
            }
        }

        private MyNorthwindDomainContext _service;
        private EntityQuery<Order> _currentQuery;
        private int _numberOfOrders = -1;
        private int _pageNumber = 0;
        private const int PageSize = 20;

        public RelayCommand SaveCommand
        {
            get;
            private set;
        }

        public RelayCommand AddCommand
        {
            get;
            private set;
        }

        public RelayCommand<Order> DeleteOrderCommand
        {
            get;
            private set;
        }

        public RelayCommand GoForwardCommand
        {
            get;
            private set;
        }

        public RelayCommand GoBackCommand
        {
            get;
            private set;
        }

        public RelayCommand PrintCommand
        {
            get;
            private set;
        }

        public IDialogService DialogService
        {
            get;
            set;
        }

        public IPrintService PrintService
        {
            get;
            set;
        }

        public void ShowMessage(string message)
        {
            if (DialogService != null)
            {
                DialogService.ShowMessage(message);
            }
        }

        private bool AskConfirmation(string message)
        {
            if (DialogService != null)
            {
                return DialogService.AskConfirmation(message);
            }
            return true;
        }

        public MainViewModel()
        {
            if (DesignerProperties.IsInDesignTool)
            {
                Orders = new ObservableCollection<Order>
                {
                    new Order { OrderID = 1234, ShipCity = "Zurich", Freight = 123 },
                    new Order { OrderID = 3456, ShipCity = "Lausanne", Freight = 42 },
                    new Order { OrderID = 4567, ShipCity = "Basel", Freight = 56 }
                };
            }
            else
            {
                SaveCommand = new RelayCommand(SaveChanges);
                AddCommand = new RelayCommand(AddOrder);
                DeleteOrderCommand = new RelayCommand<Order>(DeleteOrder);
                
                GoForwardCommand = new RelayCommand(
                    GoForward,
                    () => (_pageNumber + 1) * PageSize < _numberOfOrders);
                
                GoBackCommand = new RelayCommand(
                    GoBack,
                    () => _pageNumber > 0);

                PrintCommand = new RelayCommand(
                    () =>
                    {
                        if (PrintService != null)
                        {
                            PrintService.PrintOrders();
                        }
                    });

                _service = new MyNorthwindDomainContext();
                LoadAllOrders();
            }
        }

        public void LoadAllOrders()
        {
            IsBusy = true;
            _pageNumber = 0;
            _service.Orders.Clear();
            _currentQuery = _service.GetOrdersQuery().OrderBy(o => o.ShipCity);
            _service.GetNumberOfOrders(string.Empty, GetNumberOfOrdersCompleted, null);
            _service.Load(_currentQuery.Take(PageSize), LoadAllOrdersCompleted, null);
        }

        public void LoadAllOrdersByCity(string cityStart)
        {
            IsBusy = true;
            _pageNumber = 0;
            var query = from o in _service.GetOrdersQuery()
                        where o.ShipCity.StartsWith(cityStart)
                        orderby o.ShipCity
                        orderby o.OrderID
                        select o;

            _service.Orders.Clear();
            _currentQuery = query;
            _service.GetNumberOfOrders(cityStart, GetNumberOfOrdersCompleted, null);
            _service.Load(_currentQuery.Take(PageSize), LoadAllOrdersCompleted, null);
        }

        private void LoadAllOrdersCompleted(LoadOperation<Order> op)
        {
            if (op.HasError)
            {
                ShowError(op);
                return;
            }

            Orders = new ObservableCollection<Order>(_service.Orders);
            GoForwardCommand.RaiseCanExecuteChanged();
            GoBackCommand.RaiseCanExecuteChanged();
            IsBusy = false;
        }

        public void GetNumberOfOrdersCompleted(InvokeOperation<int> op)
        {
            _numberOfOrders = op.Value;
            GoForwardCommand.RaiseCanExecuteChanged();
        }

        public void SaveChanges()
        {
            foreach (var entity in _service.Orders)
            {
                if (entity.HasChanges
                    && entity.HasValidationErrors)
                {
                    ShowMessage("Impossible to save");
                    return;
                }
            }

            IsBusy = true;
            _service.SubmitChanges(SaveChangesCompleted, null);
        }

        private void SaveChangesCompleted(SubmitOperation op)
        {
            if (op.HasError)
            {
                if (op.EntitiesInError != null
                    && op.EntitiesInError.Count() > 0)
                {
                    var ordersError = string.Empty;
                    foreach (Order order in op.EntitiesInError)
                    {
                        ordersError
                            = string.Join(" ", ordersError, order.OrderID);
                    }

                    ShowMessage(string.Format(
                        "Refreshing because of inconsistent data:{0}",
                        ordersError));

                    op.MarkErrorAsHandled();
                    _service.Orders.Clear();
                    _service.Load(
                         _currentQuery.Take(PageSize),
                         LoadAllOrdersCompleted,
                         null);
                }
                else
                {
                    ShowError(op);
                }

                return;
            }

            IsBusy = false;
        }

        private void AddOrder()
        {
            var lastOrder = _service.Orders.OrderBy(o => o.OrderID).Last();

            var newOrder = new Order
            {
                CustomerID = "ALFKI",
                EmployeeID = 1,
                Freight = 10,
                OrderID = lastOrder.OrderID + 1
            };

            _service.Orders.Add(newOrder);
            Orders.Add(newOrder);
            _numberOfOrders++;
        }

        private void DeleteOrder(Order order)
        {
            if (!_service.Orders.Contains(order))
            {
                return;
            }

            var confirmation = string.Format(
                "Do you want to delete the order # {0}?",
                order.OrderID);

            if (AskConfirmation(confirmation))
            {
                IsBusy = true;
                _service.Orders.Remove(order);
                _service.SubmitChanges(DeleteOrderCompleted, order);
            }
        }

        private void DeleteOrderCompleted(SubmitOperation op)
        {
            if (op.HasError)
            {
                ShowError(op);
                return;
            }

            var order = op.UserState as Order;
            if (order != null)
            {
                Orders.Remove(order);
                _numberOfOrders--;
            }

            IsBusy = false;
        }

        public void GoForward()
        {
            var query = _currentQuery.Skip(++_pageNumber * PageSize)
                .Take(PageSize);

            _service.Orders.Clear();
            _service.Load(query, LoadAllOrdersCompleted, null);
        }

        public void GoBack()
        {
            var query = _currentQuery.Skip(--_pageNumber * PageSize)
                .Take(PageSize);

            _service.Orders.Clear();
            _service.Load(query, LoadAllOrdersCompleted, null);
        }

        private void ShowError(OperationBase op)
        {
            IsBusy = false;
            if (op == null
                || op.Error == null)
            {
                ShowMessage("Unknown error");
                return;
            }

            ShowMessage(
                "There was an error:"
                + Environment.NewLine
                + op.Error.Message);
            op.MarkErrorAsHandled();
        }

        public event PropertyChangedEventHandler PropertyChanged;

        public void RaisePropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
            {
                PropertyChanged(this,
                    new PropertyChangedEventArgs(propertyName));
            }
        }
    }
}
